﻿using UnityEngine;
using System.Collections.Generic;
using HarmonyLib;
using System;
using Steamworks;
using UnityEngine.AzureSky;
using System.Net.Sockets;
using System.Text;
using UnityEngine.UI;

public class CrowdControl : Mod
{
    static TcpClient client;
    static NetworkStream stream;
    public static string prefix = "[Crowd Control Support]: ";
    public const int channelID = 42;
    public const Messages messageID = (Messages)78;
    public static System.Random rand = new System.Random();
    public static ChunkPoint spawnChunk;
    public static float sharkPacify = 0;
    public static Vector4 windDirection = Vector4.zero;
    public static Tuple<CSteamID, int, string> eventRequest;
    public static List<TimedEffect> effectTimers;
    public static List<AnimalEffect> spawnedAnimals;
    public static Block cacheAttackableBlock;
    public static Transform buildPivot;
    public static BlockQuad cacheBuildPos;
    Harmony harmony;
    public void Start()
    {
        eventRequest = null;
        effectTimers = new List<TimedEffect>();
        spawnedAnimals = new List<AnimalEffect>();
        try
        {
            client = new TcpClient("127.0.0.1", 58430);
            stream = client.GetStream();
        } catch
        {
            Debug.LogError(prefix + "Initial connection failed. Please make sure the Crowd Control Client app is running then retry using the \"reconnectCrowdControl\" command");
        }
        harmony = new Harmony("com.aidanamite.CrowdControl");
        harmony.PatchAll();
        Debug.Log(prefix + "Mod has been loaded!");
    }

    public void Update()
    {
        ClientUpdate();
        float timeDif = Time.deltaTime;
        var message = RAPI.ListenForNetworkMessagesOnChannel(channelID);
        if (message != null)
        {
            if (message.message.Type == MessageType.MessageID)
            {
                var msg = message.message as Message_InitiateConnection;
                if (msg.appBuildID == MessageType.Execute)
                {
                    var info = new Message_ExecuteEvent(msg);
                    if (info.MessageID > -1)
                        eventRequest = new Tuple<CSteamID, int, string>(message.steamid, info.MessageID, "");
                    ExecuteEvent(info.EventID);
                    eventRequest = null;
                } else if (msg.appBuildID == MessageType.Executed && client.Connected)
                {
                    var info = new Message_EventExecuted(msg);
                    foreach (AnimalEffect item in spawnedAnimals)
                        if (item.id == info.MessageID)
                            item.animalId = info.ExtraData;
                    byte[] responce = Encoding.UTF8.GetBytes("{\"id\":" + info.MessageID + ",\"status\":" + (int)EffectResult.Success + ",\"message\":\"Event was executed\",\"type\":0}");
                    byte[] final = new byte[responce.Length + 1];
                    responce.CopyTo(final, 0);
                    if (client.Connected)
                        stream.Write(final, 0, final.Length);
                }
                else if (msg.appBuildID == MessageType.Wind)
                {
                    var info = new Message_Wind(msg);
                    windDirection = new Vector4((float)Math.Sin(info.Angle), 0, (float)Math.Cos(info.Angle), info.Time);
                }
                else if (msg.appBuildID == MessageType.Upgrade)
                {
                    uint ObjectIndex = new Message_Upgrade(msg).ObjectIndex;
                    BlockCreator.GetPlacedBlocks().Find((Block b) => b.ObjectIndex == ObjectIndex).Reinforced = true;
                }
            }
        }
        if (RAPI.GetLocalPlayer() == null)
        {
            if (spawnChunk != null)
                spawnChunk = null;
            if (effectTimers.Count > 0)
                effectTimers.Clear();
            if (spawnedAnimals.Count > 0)
                spawnedAnimals.Clear();
        }
        else
        {
            spawnChunk = getValidIsland();
            if (sharkPacify > 0)
                sharkPacify -= timeDif;
            for (int i = effectTimers.Count - 1;i >= 0;i--)
            {
                effectTimers[i].time += timeDif;
                if (effectTimers[i].time > 30)
                {
                    eventRequest = new Tuple<CSteamID, int, string>(new CSteamID(0L), -1, "RaftConsole");
                    effectTimers[i].reverseEffect.Execute();
                    eventRequest = null;
                    effectTimers.RemoveAt(i);
                }
            }
            foreach (AnimalEffect effect in spawnedAnimals)
                if (effect.nameTag != null)
                    effect.nameTag.UpdateFacing();
            if (windDirection.w != 0)
            {
                windDirection.w -= timeDif;
                if (windDirection.w < 0)
                    windDirection.w = 0;
            }
        }
    }

    public void OnModUnload()
    {
        if (client != null)
            client.Dispose();
        harmony.UnpatchAll(harmony.Id);
        Debug.Log(prefix + "Mod has been unloaded!");
    }

    public static void sendMessage(Message msg)
    {
        Semih_Network network = RAPI.GetLocalPlayer().Network;
        network.SendP2P(network.HostID, msg, EP2PSend.k_EP2PSendReliable, (NetworkChannel)channelID);
    }

    public static void sendReply(int extra = -1)
    {
        if (eventRequest != null)
            new Message_EventExecuted(eventRequest.Item2, extra).Message.SendP2P(eventRequest.Item1, MessageType.Channel);
    }

    public static void sendRequest(string eventID)
    {
        new Message_ExecuteEvent(events.IndexOf(eventID), eventRequest.Item2).Message.Broadcast();
    }

    public static void ClientUpdate()
    {
        if (client != null && client.Connected && stream.DataAvailable)
        {
            List<byte> bytes = new List<byte>();
            while (stream.DataAvailable)
            {
                byte data = (byte)stream.ReadByte();
                if (data == 0)
                    break;
                else
                    bytes.Add(data);
            }
            string str = Encoding.UTF8.GetString(bytes.ToArray());
            JSONObject msg;
            int msgID;
            try
            {
                byte[] responce = null;
                msg = new JSONObject(str);
                msgID = (int)msg.GetField("id").n;
                try
                {
                    string msgCode = msg.GetField("code").str;
                    if (RAPI.GetLocalPlayer() != null && events[msgCode].IsReady)
                    {
                        RequestType msgType = (RequestType)msg.GetField("type").n;
                        string strMsg = "";
                        bool flag = true;
                        if (msgType == RequestType.Start)
                        {
                            eventRequest = new Tuple<CSteamID, int, string>(new CSteamID(0L), msgID, msg.GetField("viewer").str);
                            flag = !ExecuteEvent(msgCode);
                            eventRequest = null;
                            strMsg = "Event was executed";
                        }
                        else if (msgType == RequestType.Test)
                            strMsg = "Event is executable";
                        else if (msgType == RequestType.Stop)
                        {
                            foreach (TimedEffect item in effectTimers)
                                if (item.id == msgID)
                                {
                                    item.time = 30;
                                    break;
                                }
                            foreach (AnimalEffect item in spawnedAnimals)
                                if (item.id == msgID)
                                {
                                    if (item.animalId == -1)
                                    {
                                        flag = false;
                                        responce = Encoding.UTF8.GetBytes("{\"id\":" + msgID + ",\"status\":" + (int)EffectResult.Retry + ",\"message\":\"Entity not created yet\",\"type\":0}");
                                    }
                                    else
                                    {
                                        NetworkIDManager.SendIDBehaviourDead((uint)item.animalId, typeof(AI_NetworkBehaviour), true);
                                    }
                                    break;
                                }
                            if (msgCode == "pacifyShark")
                                sharkPacify = 0;
                            if (msgCode == "changeWind")
                            {
                                windDirection.w = 0;
                                new Message_Wind(0, 0).Message.Broadcast(MessageType.Channel);
                            }
                            strMsg = "Event is not running";
                        }
                        else
                            strMsg = "Nothing happened";
                        if (flag)
                            responce = Encoding.UTF8.GetBytes("{\"id\":" + msgID + ",\"status\":" + (int)EffectResult.Success + ",\"message\":\"" + strMsg + "\",\"type\":0}");
                    }
                    else
                        responce = Encoding.UTF8.GetBytes("{\"id\":" + msgID + ",\"status\":" + (int)EffectResult.Failure + ",\"message\":\"Event is not executable\",\"type\":0}");
                }
                catch (Exception err)
                {
                    Debug.LogError("Event message failed to parse\n" + str + "\n" + err.GetType().Name + ": " + err.Message + "\n" + err.StackTrace);
                    responce = Encoding.UTF8.GetBytes("{\"id\":" + msgID + ",\"status\":" + (int)EffectResult.Failure + ",\"message\":\"Event failed due to an unknown error\",\"type\":0}");
                }
                if (responce != null)
                {
                    byte[] final = new byte[responce.Length + 1];
                    responce.CopyTo(final, 0);
                    stream.Write(final, 0, final.Length);
                }
            }
            catch (Exception err)
            {
                Debug.LogError("Event message failed to parse\n" + str + "\n" + err.GetType().Name + ": " + err.Message + "\n" + err.StackTrace);
            }
        }
    }

    [ConsoleCommand(name: "runevent", docs: "Syntax: 'runevent <eventName>' Triggers one of the Crowd Control events given the event's ID name")]
    public static string MyCommand(string[] args)
    {
        if (args.Length < 1)
            return prefix + "Not enough parameters";
        eventRequest = new Tuple<CSteamID, int, string>(new CSteamID(0L), -1, "RaftConsole");
        ExecuteEvent(args[0]);
        eventRequest = null;
        return prefix + "Execution attempt complete";
    }

    [ConsoleCommand(name: "stopevent", docs: "Syntax: 'stopevent <id>' Stops an event given its id")]
    public static string MyCommand2(string[] args)
    {
        if (args.Length > 1)
            return prefix + "Too many parameters";
        if (args.Length < 1)
            return prefix + "Not enough parameters";
        int port;
        if (!int.TryParse(args[0], out port))
            return prefix + "Could not parse \"" + args[0] + "\" as a number";
        try
        {
            foreach (AnimalEffect item in spawnedAnimals)
                if (item.id == port)
                {
                    if (item.animalId == -1)
                        return prefix + "Event has not occured yet";
                    else
                    {
                        NetworkIDManager.SendIDBehaviourDead((uint)item.animalId, typeof(AI_NetworkBehaviour), true);
                    }
                    return prefix + "Successfully stopped event " + args[0];
                }
            return prefix + "Did not find event " + args[0];
        }
        catch (Exception err)
        {
            return prefix + "Something went wrong\n" + err.GetType().Name + ": " + err.Message + "\n" + err.StackTrace;
        }
    }

    [ConsoleCommand(name: "reconnectCrowdControl", docs: "Syntax: 'reconnectCrowdControl' Attempts to connect the crowd control client to the stored address")]
    public static string MyCommand3(string[] args)
    {
        try
        {
            client = new TcpClient("127.0.0.1", 58430);
            stream = client.GetStream();
            return prefix + "Successfully connected";
        }
        catch
        {
            return prefix + "Connection failed. Please make sure the Crowd Control Client app is running then retry";
        }
    }

    public static List<string> BadItems = new List<string>
    {
        "DevSpear",
        "DevHat",
        "ThrowableAnchor",
        "BeachBall",
        "Block_FoundationArmor",
        "Repair",
        "Barrel",
        "DropItem",
        "Blueprint",
        "Seed_Grass",
        "Block_Upgrade_Thatch",
        "Block_Upgrade_Wood",
        "Block_Upgrade_Tier3"
    };

    public static List<Item_Base> VisionHats = new List<Item_Base>
    {
        ItemManager.GetItemByIndex(306),
        ItemManager.GetItemByIndex(205),
        ItemManager.GetItemByIndex(127),
        ItemManager.GetItemByIndex(271)
    };

    static class readyTypes
    {
        public static Func<bool> None = delegate { return true; };
        public static Func<bool> PlayerIsHoldingItem = delegate { return RAPI.GetLocalPlayer().Inventory.GetSelectedHotbarSlot().HasValidItemInstance(); };
        public static Func<bool> PlayerHasAnyItem = delegate { foreach (Slot slot in RAPI.GetLocalPlayer().Inventory.allSlots) if (slot.active && slot.HasValidItemInstance() && !slot.locked) return true; return false; };
        public static Func<bool> PlayerIsNotDead = delegate { return !RAPI.GetLocalPlayer().Stats.IsDead; };
        public static Func<bool> RaftIsAttackable = delegate { return FindObjectsOfType<Block_Foundation>().Length > 4 && getTargetableBlock(); };
        public static Func<bool> HasNearbyIsland = delegate { return spawnChunk != null; };
        public static Func<bool> PlayerHasAnyTool = delegate { foreach (Slot slot in RAPI.GetLocalPlayer().Inventory.allSlots) if (slot.HasToolItem()) return true; return false; };
        public static Func<bool> CanUpgradeRaft = delegate { RaftBounds bounds = FindObjectOfType<RaftBounds>(); if (bounds == null) return false; foreach (Block block in bounds.walkableBlocks) if (block is Block_Foundation && !block.Reinforced) return true;return false; };
        public static Func<bool> PlayerCanEquipHat = delegate { foreach (Slot_Equip slot in RAPI.GetLocalPlayer().Inventory.equipSlots) if (!slot.HasValidItemInstance() || slot.itemInstance.settings_equipment.EquipType == EquipSlotType.Head) return true; return false; };
        public static Func<bool> SharkIsAlive = delegate { foreach (AI_NetworkBehaviour entity in FindObjectsOfType<AI_NetworkBehaviour>()) if (entity is AI_NetworkBehavior_Shark && !entity.networkEntity.IsDead) return true; return false; };
        public static Func<bool> GravityEffectActive = delegate { foreach (TimedEffect effect in effectTimers) if (effect.reverseEffect.id.StartsWith("gravity")) return false; return true; };
        public static Func<bool> SpeedEffectActive = delegate { foreach (TimedEffect effect in effectTimers) if (effect.reverseEffect.id.StartsWith("speed")) return false; return true; };
        public static Func<bool> RaftSpeedEffectActive = delegate { foreach (TimedEffect effect in effectTimers) if (effect.reverseEffect.id.StartsWith("raftSpeed")) return false; return true; };
        public static Func<bool> WindEffectActive = delegate { return windDirection.w == 0; };
        public static Func<bool> PacifyEffectActive = delegate { return sharkPacify == 0; };
        public static Func<bool> CanDropAnchors = CanDropAnchor;
        public static Func<bool> CanRaiseAnchors = CanRaiseAnchor;
        public static Func<bool> CanTurnOnEngines = CanEngineOn;
        public static Func<bool> CanTurnOffEngines = CanEngineOff;
        public static Func<bool> HasEngines = delegate { return FindObjectOfType<MotorWheel>() != null || FindObjectOfType<Sail>() != null; };
    };

    public static EffectList events = new EffectList
    {
        new Effect("spawnShark", spawnShark,readyTypes.None,true),
        new Effect("spawnItem",delegate { RAPI.GetLocalPlayer().giveItem();},readyTypes.None,false),
        new Effect("spawnItem_food",delegate { RAPI.GetLocalPlayer().giveItem("food"); },readyTypes.None,false),
        new Effect("spawnItem_drink",delegate { RAPI.GetLocalPlayer().giveItem("drink"); },readyTypes.None,false),
        new Effect("spawnItem_edible",delegate { RAPI.GetLocalPlayer().giveItem("edible"); },readyTypes.None,false),
        new Effect("spawnItem_smelted",delegate { RAPI.GetLocalPlayer().giveItem("smelted"); },readyTypes.None,false),
        new Effect("spawnItem_block",delegate { RAPI.GetLocalPlayer().giveItem("block"); },readyTypes.None,false),
        new Effect("spawnItem_equip",delegate { RAPI.GetLocalPlayer().giveItem("equip"); },readyTypes.None,false),
        new Effect("spawnItem_copy",delegate { RAPI.GetLocalPlayer().giveItem("copy"); },readyTypes.None,false),
        new Effect("spawnItem_weapon",delegate { RAPI.GetLocalPlayer().giveItem("weapon"); },readyTypes.None,false),
        new Effect("spawnItem_resource",delegate { RAPI.GetLocalPlayer().giveItem("resource"); },readyTypes.None,false),
        new Effect("dropItem",delegate { RAPI.GetLocalPlayer().DropCurrentItem(); },readyTypes.PlayerIsHoldingItem,false),
        new Effect("dropRandomItem",delegate { RAPI.GetLocalPlayer().DropRandomItem(); },readyTypes.PlayerHasAnyItem,false),
        new Effect("skipTime",skipTime,readyTypes.None,true),
        new Effect("healthUp",delegate { changeHealth(10); },readyTypes.PlayerIsNotDead,false),
        new Effect("healthUp2",delegate { changeHealth(25); },readyTypes.PlayerIsNotDead,false),
        new Effect("healthDown",delegate { changeHealth(-10); },readyTypes.PlayerIsNotDead,false),
        new Effect("healthDown2",delegate { changeHealth(-25); },readyTypes.PlayerIsNotDead,false),
        new Effect("foodUp",delegate { changeHunger(10); },readyTypes.PlayerIsNotDead,false),
        new Effect("foodUp2",delegate { changeHunger(25); },readyTypes.PlayerIsNotDead,false),
        new Effect("foodDown",delegate { changeHunger(-10); },readyTypes.PlayerIsNotDead,false),
        new Effect("foodDown2",delegate { changeHunger(-25); },readyTypes.PlayerIsNotDead,false),
        new Effect("waterUp",delegate { changeThirst(10); },readyTypes.PlayerIsNotDead,false),
        new Effect("waterUp2",delegate { changeThirst(25); },readyTypes.PlayerIsNotDead,false),
        new Effect("waterDown",delegate { changeThirst(-10); },readyTypes.PlayerIsNotDead,false),
        new Effect("waterDown2",delegate { changeThirst(-25); },readyTypes.PlayerIsNotDead,false),
        new Effect("speedUp",delegate { changeSpeed(true); },readyTypes.SpeedEffectActive,false),
        new Effect("speedUp2",delegate { changeSpeed(true); changeSpeed(true); },readyTypes.SpeedEffectActive,false),
        new Effect("speedDown",delegate { changeSpeed(false); },readyTypes.SpeedEffectActive,false),
        new Effect("speedDown2",delegate { changeSpeed(false); changeSpeed(false); },readyTypes.SpeedEffectActive,false),
        new Effect("gravityUp",delegate { changeGravity(true, eventRequest == null ? -1 : eventRequest.Item2); },readyTypes.GravityEffectActive,false),
        new Effect("gravityDown",delegate { changeGravity(false, eventRequest == null ? -1 : eventRequest.Item2); },readyTypes.GravityEffectActive,false),
        new Effect("raftSpeedUp",delegate { changeRaftSpeed(true); },readyTypes.RaftSpeedEffectActive,true),
        new Effect("raftSpeedDown",delegate { changeRaftSpeed(false); },readyTypes.RaftSpeedEffectActive,true),
        new Effect("phantomShark",delegate { BlockCreator.RemoveBlockNetwork(cacheAttackableBlock, null); },readyTypes.RaftIsAttackable,false),
        new Effect("spawnBear",delegate { trySpawnAnimal(AI_NetworkBehaviourType.Bear); },readyTypes.HasNearbyIsland,true),
        new Effect("spawnMamaBear",delegate { trySpawnAnimal(AI_NetworkBehaviourType.MamaBear); },readyTypes.HasNearbyIsland,true),
        new Effect("spawnStoneBird",delegate { trySpawnAnimal(AI_NetworkBehaviourType.StoneBird); },readyTypes.HasNearbyIsland,true),
        new Effect("spawnStoneBird_Caravan",delegate { trySpawnAnimal(AI_NetworkBehaviourType.StoneBird_Caravan); },readyTypes.HasNearbyIsland,true),
        new Effect("spawnBoar",delegate { trySpawnAnimal(AI_NetworkBehaviourType.Boar); },readyTypes.HasNearbyIsland,true),
        new Effect("spawnButlerBot",delegate { trySpawnAnimal(AI_NetworkBehaviourType.ButlerBot); },readyTypes.HasNearbyIsland,true),
        new Effect("spawnPig",delegate { trySpawnAnimal(AI_NetworkBehaviourType.Pig); },readyTypes.HasNearbyIsland,true),
        new Effect("pacifyShark",pacifyShark,readyTypes.PacifyEffectActive,true),
        new Effect("repairRandomTool",delegate { RAPI.GetLocalPlayer().RepairRandomItem(); },readyTypes.PlayerHasAnyTool,false),
        new Effect("damageRandomTool",delegate { RAPI.GetLocalPlayer().DamageRandomItem(); },readyTypes.PlayerHasAnyTool,false),
        new Effect("teleportToRaft", teleportToRaft, readyTypes.None, false),
        new Effect("upgradeRaft", upgradeRaft, readyTypes.CanUpgradeRaft, false),
        new Effect("changeWind", randomizeWind, readyTypes.WindEffectActive, false),
        new Effect("killRandomShark", killRandomShark, readyTypes.SharkIsAlive, true),
        new Effect("changeHat", changeHat, readyTypes.PlayerCanEquipHat, false),
        new Effect("raiseAnchors", RaiseAllAnchors, readyTypes.CanRaiseAnchors, false),
        new Effect("dropAnchors", DropAllAnchors, readyTypes.CanDropAnchors, false),
        new Effect("enginesOn", AllEnginesOn, readyTypes.CanTurnOnEngines, false),
        new Effect("enginesOff", AllEnginesOff, readyTypes.CanTurnOffEngines, false),
        new Effect("changeDirections", ChangeEngineDirections, readyTypes.HasEngines, false)
    };

    public static bool ExecuteEvent(int id)
    {
        if (events.Count < id || id < 0)
            Debug.LogError("Did not find event for numeric id " + id);
        Effect e = events[id];
        if (eventRequest != null && isTimedEvent(e.id))
            effectTimers.Add(new TimedEffect(eventRequest.Item2, reverseEvent(e.id)));
        return e.Execute();
    }
    public static bool ExecuteEvent(string id)
    {
        Effect e = events[id];
        if (e == null)
            Debug.LogError("Did not find event for string id " + id);
        if (eventRequest != null && isTimedEvent(e.id))
            effectTimers.Add(new TimedEffect(eventRequest.Item2, reverseEvent(e.id)));
        return e.Execute();
    }

    public static bool isTimedEvent(string id)
    {
        return id.StartsWith("gravity") || id.StartsWith("speed") || id.StartsWith("raftSpeed");
    }

    public static Effect reverseEvent(string id)
    {
        return events[id.Contains("Up") ? id.Replace("Up", "Down") : id.Replace("Down", "Up")];
    }

    public static void spawnShark()
    {
        if (Semih_Network.IsHost)
        {
            Network_Host_Entities entity = ComponentManager<Network_Host_Entities>.Value;
            AI_NetworkBehaviour anb = entity.CreateAINetworkBehaviour(AI_NetworkBehaviourType.Shark, entity.GetSharkSpawnPosition());
            if (eventRequest.Item1.m_SteamID == 0)
                spawnedAnimals.Add(new AnimalEffect(eventRequest.Item2, eventRequest.Item3, (int)anb.ObjectIndex));
            sendReply((int)anb.ObjectIndex);
        }
        else
        {
            if (eventRequest.Item1.m_SteamID == 0)
                spawnedAnimals.Add(new AnimalEffect(eventRequest.Item2, eventRequest.Item3, -1));
            sendRequest("spawnShark");
        }
    }

    public static void skipTime()
    {
        if (Semih_Network.IsHost)
        {
            AzureSkyTimeOfDayComponent sc = ComponentManager<AzureSkyController>.Value.timeOfDay;
            if (sc.hour > 12)
            {
                sc.hour -= 12;
                sc.JumpToNextDay();
            }
            else
                sc.hour += 12;
            sendReply();
        }
        else
            sendRequest("skipTime");
    }

    public static void changeHealth(float amount)
    {
        RAPI.GetLocalPlayer().Stats.Damage(-amount, RAPI.GetLocalPlayer().PersonController.transform.position, Vector3.up, EntityType.Environment);
    }

    public static void changeHunger(float amount)
    {
        RAPI.GetLocalPlayer().Stats.stat_hunger.Normal.Value += amount;
    }

    public static void changeThirst(float amount)
    {
        RAPI.GetLocalPlayer().Stats.stat_thirst.Normal.Value += amount;
    }

    public static void changeSpeed(bool increase)
    {
        PersonController control = RAPI.GetLocalPlayer().PersonController;
        if (increase)
        {
            control.normalSpeed *= 1.5f;
            control.sprintSpeed *= 1.5f;
            control.swimSpeed *= 1.5f;
        }
        else
        {
            control.normalSpeed /= 1.5f;
            control.sprintSpeed /= 1.5f;
            control.swimSpeed /= 1.5f;
        }
    }

    public static void changeGravity(bool increase, int eventID = -1)
    {
        if (increase)
            RAPI.GetLocalPlayer().PersonController.gravity *= 1.5f;
        else
            RAPI.GetLocalPlayer().PersonController.gravity /= 1.5f;
    }

    public static void changeRaftSpeed(bool increase)
    {
        if (Semih_Network.IsHost)
        {
            if (increase)
                ComponentManager<Raft>.Value.maxSpeed *= 1.2f;
            else
                ComponentManager<Raft>.Value.maxSpeed /= 1.2f;
            sendReply();
        }
        else
            sendRequest(increase ? "raftSpeedUp" : "raftSpeedDown");
    }

    public static bool getTargetableBlock()
    {

        RaftBounds bounds = FindObjectOfType<RaftBounds>();
        List<Block> blocks = new List<Block>();
        List<AngleRange> ranges = new List<AngleRange>();
        foreach (Block block in bounds.walkableBlocks)
            if (block is Block_Foundation)
            {
                blocks.Add(block);
                ranges.Add(new AngleRange());
            }
        for (int i = 0; i < blocks.Count; i++)
            for (int j = 0; j < blocks.Count; j++)
            {
                float dist = blocks[i].transform.position.DistanceXZ(blocks[j].transform.position) / 2;
                if (j != i && dist < 5)
                {
                    double vary = Math.Acos(dist / 5);
                    double main = Math.Acos(((double)blocks[i].transform.position.x - blocks[j].transform.position.x) / (dist * 2)) * (blocks[i].transform.position.z < blocks[j].transform.position.z ? -1 : 1);
                    ranges[i] -= new AngleRange(main - vary, main + vary);
                    ranges[j] -= new AngleRange(main - vary + Math.PI, main + vary + Math.PI);
                }
            }
        List<Block> tmp = new List<Block>();
        for (int i = 0; i < blocks.Count; i++)
            if (!ranges[i].Empty)
            {
                tmp.Add(blocks[i]);
            }
        if (tmp.Count == 0)
            return false;
        cacheAttackableBlock = tmp[(int)(rand.NextDouble() * tmp.Count)];
        return true;
    }

    public static ChunkPoint getValidIsland()
    {
        ChunkManager manager = ComponentManager<ChunkManager>.Value;
        ChunkPoint nearestValidChunk = null;
        float dist = float.MaxValue;
        Vector3 player = RAPI.GetLocalPlayer().PersonController.transform.position;
        List<ChunkPoint> chunkPointsFromPointType = manager.GetAllChunkPointsList();
        foreach (ChunkPoint chunk in chunkPointsFromPointType)
            if (chunk.spawner != null && chunk.spawner.ChunkPointType != ChunkPointType.None)
            {
                float newDist = player.DistanceXZ(chunk.worldPosition);
                if (newDist < dist)
                {
                    nearestValidChunk = chunk;
                    dist = newDist;
                }
            }
        if (dist > 200)
            return null;
        return nearestValidChunk;
    }

    public static AI_NetworkBehaviour spawnAnimal(AI_NetworkBehaviourType animalType)
    {
        Vector3 spawnPos = Vector3.zero;
        bool flag = false;
        float start = Time.time;
        while (!flag)
        {
            spawnPos = spawnChunk.worldPosition.XZOnly() + new Vector3((float)(rand.NextDouble() - 0.5) * ChunkManager.ChunkSize, 200, (float)(rand.NextDouble() - 0.5) * ChunkManager.ChunkSize);
            foreach (RaycastHit hit in Physics.RaycastAll(spawnChunk.worldPosition.XZOnly() + new Vector3((float)(rand.NextDouble() - 0.5) * ChunkManager.ChunkSize, 200, (float)(rand.NextDouble() - 0.5) * ChunkManager.ChunkSize), Vector3.down, 500))
            {
                if (spawnPos.y > hit.point.y && hit.point.y > 1 && !hit.collider.transform.ParentedToRaft() && hit.collider.GetComponent<PickupItem>() == null)
                {
                    flag = true;
                    spawnPos = hit.point;
                }
            }
            if (Time.time > start + 10)
                throw new OperationCanceledException("Failed to find animal spawn location in time.");
        }
        return ComponentManager<Network_Host_Entities>.Value.CreateAINetworkBehaviour(animalType, spawnPos);
    }

    public static void trySpawnAnimal(AI_NetworkBehaviourType animal)
    {
        if (Semih_Network.IsHost)
        {
            AI_NetworkBehaviour anb = spawnAnimal(animal);
            if (eventRequest.Item1.m_SteamID == 0)
                spawnedAnimals.Add(new AnimalEffect(eventRequest.Item2, eventRequest.Item3, (int)anb.ObjectIndex));
            sendReply((int)anb.ObjectIndex);
        }
        else
        {
            spawnedAnimals.Add(new AnimalEffect(eventRequest.Item2, eventRequest.Item3, -1));
            sendRequest("spawn" + animal.ToString());
        }
    }

    public static void pacifyShark()
    {
        if (Semih_Network.IsHost)
        {
            sharkPacify = 60;
            sendReply();
        }
        else
            sendRequest("pacifyShark");
    }

    public static void teleportToRaft()
    {
        Network_Player player = RAPI.GetLocalPlayer();
        RaftBounds bounds = FindObjectOfType<RaftBounds>();
        Vector3 pos = bounds.walkableBlocks[(int)(rand.NextDouble() * bounds.walkableBlocks.Count)].transform.position + new Vector3(0, 1, 0);
        ComponentManager<Semih_Network>.Value.RPC(new Message_Teleport(Messages.Teleport, player.Network.NetworkIDManager, player, pos), Target.All, EP2PSend.k_EP2PSendReliable, NetworkChannel.Channel_Game);
    }

    public static void upgradeRaft()
    {
        Network_Player player = RAPI.GetLocalPlayer();
        List<Block> blocks = new List<Block>();
        RaftBounds bounds = FindObjectOfType<RaftBounds>();
        foreach (Block block in bounds.walkableBlocks)
            if (block is Block_Foundation && !block.Reinforced)
                blocks.Add(block);
        Block block_Foundation = blocks[(int)(rand.NextDouble() * blocks.Count)];
        block_Foundation.Reinforced = true;
        new Message_Upgrade(block_Foundation.ObjectIndex).Message.Broadcast(MessageType.Channel);
    }

    public static void randomizeWind()
    {
        double angle = rand.NextDouble() * 2 * Math.PI;
        windDirection = new Vector4((float)Math.Sin(angle),0,(float)Math.Cos(angle),120);
        new Message_Wind((float)angle, 120).Message.Broadcast(MessageType.Channel);
    }

    public static void changeHat()
    {
        Network_Player player = RAPI.GetLocalPlayer();
        List<Slot_Equip> empty = new List<Slot_Equip>();
        List<Slot_Equip> hats = new List<Slot_Equip>();
        foreach (Slot_Equip slot in player.Inventory.equipSlots)
            if (!slot.HasValidItemInstance())
                empty.Add(slot);
            else if (slot.itemInstance.settings_equipment.EquipType == EquipSlotType.Head)
                hats.Add(slot);
        if (hats.Count == 0)
            empty[0].SetItem(VisionHats[(int)(rand.NextDouble() * VisionHats.Count)], 1);
        else
        {
            ItemInstance item = hats[0].itemInstance;
            hats[0].SetItem(VisionHats[(int)(rand.NextDouble() * VisionHats.Count)], 1);
            player.Inventory.AddItem(item);
        }
    }

    public static void killRandomShark()
    {
        List<Network_Entity> entities = new List<Network_Entity>();
        foreach (AI_NetworkBehaviour entity in FindObjectsOfType<AI_NetworkBehaviour>())
            if (entity is AI_NetworkBehavior_Shark && !entity.networkEntity.IsDead)
                entities.Add(entity.networkEntity);
        Network_Entity target = entities[(int)(rand.NextDouble() * entities.Count)];
        Semih_Network network = target.Network;
        Message_NetworkEntity_Damage message = new Message_NetworkEntity_Damage(Messages.DamageEntity, network.NetworkIDManager, ComponentManager<Network_Host_Entities>.Value.ObjectIndex, target.ObjectIndex, target.stat_health.Value, target.transform.position, Vector3.up, EntityType.Environment, null);
        if (Semih_Network.IsHost)
        {
            target.Damage(message.damage, message.HitPosition, message.HitNormal, message.damageInflictorEntityType, null);
            RAPI.SendNetworkMessage(message);
        }
        else
            network.SendP2P(network.HostID, message, EP2PSend.k_EP2PSendReliable, NetworkChannel.Channel_Game);
    }

    public static GameObject CreateText(Transform canvas_transform, float x, float y, string text_to_print, int font_size, Color text_color, float width, float height, Font font, string name = "Text")
    {
        GameObject UItextGO = new GameObject("Text");
        UItextGO.transform.SetParent(canvas_transform, false);
        RectTransform trans = UItextGO.AddComponent<RectTransform>();
        trans.sizeDelta = new Vector2(width, height);
        trans.anchoredPosition = new Vector2(x, y);
        Text text = UItextGO.AddComponent<Text>();
        text.text = text_to_print;
        text.font = font;
        text.fontSize = font_size;
        text.color = text_color;
        text.name = name;
        Shadow shadow = UItextGO.AddComponent<Shadow>();
        shadow.effectColor = new Color();
        return UItextGO;
    }
    public static void AddTextShadow(GameObject textObject, Color shadowColor, Vector2 shadowOffset)
    {
        Shadow shadow = textObject.AddComponent<Shadow>();
        shadow.effectColor = shadowColor;
        shadow.effectDistance = shadowOffset;
    }
    public static void CopyTextShadow(GameObject textObject, GameObject shadowSource)
    {
        Shadow sourcesShadow = shadowSource.GetComponent<Shadow>();
        if (sourcesShadow == null)
            sourcesShadow = shadowSource.GetComponentInChildren<Shadow>();
        AddTextShadow(textObject, sourcesShadow.effectColor, sourcesShadow.effectDistance);
    }

    public static void DropAllAnchors()
    {
        foreach (var anchor in GameObject.FindObjectsOfType<Anchor_Stationary>())
            if (anchor.CanUse && !anchor.AtBottom)
                anchor.UseAnchor();
    }

    public static bool CanDropAnchor()
    {
        foreach (var anchor in GameObject.FindObjectsOfType<Anchor_Stationary>())
            if (anchor.CanUse && !anchor.AtBottom)
                return true;
        return false;
    }

    public static void RaiseAllAnchors()
    {
        foreach (var anchor in GameObject.FindObjectsOfType<Anchor_Stationary>())
            if (anchor.CanUse && anchor.AtBottom)
                anchor.UseAnchor();
    }

    public static bool CanRaiseAnchor()
    {
        foreach (var anchor in GameObject.FindObjectsOfType<Anchor_Stationary>())
            if (anchor.CanUse && anchor.AtBottom)
                return true;
        return false;
    }

    public static void AllEnginesOn()
    {
        foreach (var engine in RaftVelocityManager.Motors)
            if (!engine.engineSwitchOn)
                engine.OnButtonPressed_ToggleEngine();
        foreach (var sail in Sail.AllSails)
            if (!sail.open)
                sail.Toggle();
    }

    public static bool CanEngineOn()
    {
        foreach (var engine in RaftVelocityManager.Motors)
            if (!engine.engineSwitchOn)
                return true;
        foreach (var sail in Sail.AllSails)
            if (!sail.open)
                return true;
        return false;
    }

    public static void AllEnginesOff()
    {
        foreach (var engine in RaftVelocityManager.Motors)
            if (engine.engineSwitchOn)
                engine.OnButtonPressed_ToggleEngine();
        foreach (var sail in Sail.AllSails)
            if (sail.open)
                sail.Toggle();
    }

    public static bool CanEngineOff()
    {
        foreach (var engine in RaftVelocityManager.Motors)
            if (engine.engineSwitchOn)
                return true;
        foreach (var sail in Sail.AllSails)
            if (sail.open)
                return true;
        return false;
    }

    public static void ChangeEngineDirections()
    {
        foreach (var engine in RaftVelocityManager.Motors)
            engine.OnButtonPressed_ChangeDirection();
        foreach (var sail in Sail.AllSails)
            sail.Rotate(180);
    }
}

public static class ExtentionMethods
{
    public static void giveItem(this Network_Player player, string spawnType = "all")
    {
        List<Item_Base> items = player.getSpawnItems(spawnType);
        Item_Base item = items[CrowdControl.rand.Next(items.Count - 1)];
        Debug.Log(item.UniqueIndex);
        player.Inventory.AddItem(new ItemInstance(item, (spawnType == "resource") ? 10 : 1, item.MaxUses));
    }

    public static List<Item_Base> getSpawnItems(this Network_Player player, string spawnType = "all")
    {
        List<Item_Base> items = new List<Item_Base>();
        switch (spawnType)
        {
            case "food":
                foreach (Item_Base item in ItemManager.GetAllItems())
                    if ((item.settings_consumeable.HungerYield > 0 || item.settings_consumeable.BonusHungerYield > 0) && !CrowdControl.BadItems.Contains(item.UniqueName))
                        items.Add(item);
                break;
            case "drink":
                foreach (Item_Base item in ItemManager.GetAllItems())
                    if (item.settings_consumeable.ThirstYield > 0 && !CrowdControl.BadItems.Contains(item.UniqueName))
                        items.Add(item);
                break;
            case "edible":
                foreach (Item_Base item in ItemManager.GetAllItems())
                    if ((item.settings_consumeable.HungerYield > 0 || item.settings_consumeable.BonusHungerYield > 0 || item.settings_consumeable.ThirstYield > 0) && !CrowdControl.BadItems.Contains(item.UniqueName))
                        items.Add(item);
                break;
            case "smelted":
                foreach (Item_Base item in ItemManager.GetAllItems())
                    if (item.settings_cookable.CookingResult != null && item.settings_cookable.CookingResult.item != null && !CrowdControl.BadItems.Contains(item.UniqueName))
                        items.Add(item.settings_cookable.CookingResult.item);
                break;
            case "block":
                foreach (Item_Base item in ItemManager.GetAllItems())
                    if (item.settings_buildable.Placeable && !CrowdControl.BadItems.Contains(item.UniqueName))
                        items.Add(item);
                break;
            case "equip":
                foreach (Item_Base item in ItemManager.GetAllItems())
                    if (item.settings_equipment.EquipType != EquipSlotType.None && !CrowdControl.BadItems.Contains(item.UniqueName))
                        items.Add(item);
                break;
            case "copy":
                foreach (Slot item in player.Inventory.allSlots)
                    if (item.active && item.HasValidItemInstance() && !CrowdControl.BadItems.Contains(item.GetItemBase().UniqueName))
                        items.AddUniqueOnly(item.GetItemBase());
                break;
            case "weapon":
                foreach (ItemConnection connect in player.PlayerItemManager.useItemController.allConnections)
                    if ((connect.obj.GetComponent<MeleeWeapon>() != null || connect.obj.GetComponent<ThrowableComponent>() || connect.obj.GetComponent<Throwable>()) && !CrowdControl.BadItems.Contains(connect.inventoryItem.UniqueName))
                        items.Add(connect.inventoryItem);
                break;
            case "resource":
                foreach (Item_Base item in ItemManager.GetAllItems())
                    if (item.settings_consumeable.FoodType == FoodType.None && item.settings_cookable.CookingResult.item == null && item.settings_recipe.CraftingCategory == CraftingCategory.Nothing && !item.settings_buildable.Placeable && !item.settings_buildable.HasBuildablePrefabs && item.settings_equipment.EquipType == EquipSlotType.None && !item.settings_recipe.IsBlueprint && !CrowdControl.BadItems.Contains(item.UniqueName))
                        items.Add(item);
                break;
            case "all":
                foreach (Item_Base item in ItemManager.GetAllItems())
                    if (item.settings_buildable.Placeable == item.settings_buildable.HasBuildablePrefabs && !CrowdControl.BadItems.Contains(item.UniqueName) && !item.settings_recipe.IsBlueprint)
                        items.Add(item);
                break;
        }
        if (items.Count == 0)
            items.Add(ItemManager.GetItemByName("Plank"));
        return items;
    }

    public static void DropRandomItem(this Network_Player player)
    {
        List<Slot> items = new List<Slot>();
        foreach (Slot item in player.Inventory.allSlots)
            if (item.active && item.HasValidItemInstance() && !item.locked)
                items.Add(item);
        if (player.Inventory.GetSelectedHotbarItem() != null && player.Inventory.GetSelectedHotbarItem().UniqueName != "Hammer")
            ComponentManager<CanvasHelper>.Value.CloseMenu(MenuType.BuildMenu);
        player.Inventory.DropItem(items[CrowdControl.rand.Next(items.Count - 1)]);
    }

    public static void DropCurrentItem(this Network_Player player)
    {
        ComponentManager<CanvasHelper>.Value.CloseMenu(MenuType.BuildMenu);
        RAPI.GetLocalPlayer().Inventory.DropCurrentItem();
    }

    public static void RepairRandomItem(this Network_Player player)
    {
        List<Slot> items = new List<Slot>();
        foreach (Slot item in player.Inventory.allSlots)
            if (item.active && item.HasValidItemInstance() && item.itemInstance.settings_consumeable.FoodType == FoodType.None && item.itemInstance.BaseItemMaxUses > 1)
                items.Add(item);
        Slot Target = items[CrowdControl.rand.Next(items.Count - 1)];
        Target.IncrementUses(Target.itemInstance.BaseItemMaxUses - Target.itemInstance.Uses);
    }

    public static void DamageRandomItem(this Network_Player player)
    {
        List<Slot> items = new List<Slot>();
        foreach (Slot item in player.Inventory.allSlots)
            if (item.active && item.HasValidItemInstance() && item.itemInstance.settings_consumeable.FoodType == FoodType.None && item.itemInstance.BaseItemMaxUses > 1)
                items.Add(item);
        Slot Target = items[CrowdControl.rand.Next(items.Count - 1)];
        Target.IncrementUses(-(int)Math.Ceiling(Target.itemInstance.BaseItemMaxUses / 10f));
    }

    public static bool HasToolItem(this Slot slot)
    {
        return slot.active && slot.HasValidItemInstance() && slot.itemInstance.settings_consumeable.FoodType == FoodType.None && slot.itemInstance.BaseItemMaxUses > 1;
    }

    public static void UseAnchor(this Anchor_Stationary anchor)
    {
        new Message_NetworkBehaviour(Messages.StationaryAnchorUse, anchor).Broadcast();
        if (Semih_Network.IsHost)
            Traverse.Create(anchor).Method("Use").GetValue();
    }

    public static void Toggle(this Sail sail)
    {
        new Message_NetworkBehaviour(sail.open ? Messages.Sail_Close : Messages.Sail_Open, sail).Broadcast();
        if (Semih_Network.IsHost)
        {
            if (sail.open)
                Traverse.Create(sail).Method("Close").GetValue();
            else
                sail.Open();
        }
    }

    public static void Rotate(this Sail sail, float amount)
    {
        if (Semih_Network.IsHost)
            Traverse.Create(sail).Method("Rotate", new object[] { amount }).GetValue();
        else
            new Message_Sail_Rotate(Messages.Sail_Rotate, sail, amount).Broadcast();
    }

    public static void Broadcast(this Message message, NetworkChannel channel = 0)
    {
        var network = ComponentManager<Semih_Network>.Value;
        if (Semih_Network.IsHost)
            network.RPC(message, Target.Other, EP2PSend.k_EP2PSendReliable, channel);
        else
            network.SendP2P(network.HostID, message, EP2PSend.k_EP2PSendReliable, channel);
    }

    public static void SendP2P(this Message message, CSteamID steamID, NetworkChannel channel = 0)
    {
        ComponentManager<Semih_Network>.Value.SendP2P(steamID, message, EP2PSend.k_EP2PSendReliable, channel);
    }

    public static string String(this byte[] bytes, int length = -1, int offset = 0)
    {
        string str = "";
        if (length == -1)
            length = (bytes.Length - offset) / 2;
        while (str.Length < length)
        {
            str += BitConverter.ToChar(bytes, offset + str.Length * 2);
        }
        return str;

    }
    public static string String(this List<byte> bytes) => bytes.ToArray().String();
    public static byte[] Bytes(this string str)
    {
        var data = new List<byte>();
        foreach (char chr in str)
            data.AddRange(BitConverter.GetBytes(chr));
        return data.ToArray();
    }
    public static int Integer(this byte[] bytes, int offset = 0) => BitConverter.ToInt32(bytes, offset);
    public static uint UInteger(this byte[] bytes, int offset = 0) => BitConverter.ToUInt32(bytes, offset);
    public static float Float(this byte[] bytes, int offset = 0) => BitConverter.ToSingle(bytes, offset);
    public static Vector3 Vector3(this byte[] bytes, int offset = 0) => new Vector3(bytes.Float(offset), bytes.Float(offset + 4), bytes.Float(offset + 8));
    public static byte[] Bytes(this int value) => BitConverter.GetBytes(value);
    public static byte[] Bytes(this uint value) => BitConverter.GetBytes(value);
    public static byte[] Bytes(this float value) => BitConverter.GetBytes(value);
    public static byte[] Bytes(this Vector3 value)
    {
        var data = new byte[12];
        value.x.Bytes().CopyTo(data, 0);
        value.y.Bytes().CopyTo(data, 4);
        value.z.Bytes().CopyTo(data, 8);
        return data;
    }
}
public struct MessageType
{
    public const int ChannelID = 42;
    public const NetworkChannel Channel = (NetworkChannel)ChannelID;
    public const Messages MessageID = (Messages)78;
    public const int Execute = 0;
    public const int Executed = 1;
    public const int Wind = 2;
    public const int Upgrade = 3;
}

class Message_ExecuteEvent
{
    public Message_InitiateConnection Message
    {
        get
        {
            var data = new List<byte>();
            data.AddRange(eventID.Bytes());
            data.AddRange(messageID.Bytes());
            return new Message_InitiateConnection(MessageType.MessageID, MessageType.Execute, data.String());
        }
    }
    public int EventID { get { return eventID; } }
    public int MessageID { get { return messageID; } }
    int eventID;
    int messageID;
    public Message_ExecuteEvent(int eventId, int messageId)
    {
        eventID = eventId;
        messageID = messageId;
    }
    public Message_ExecuteEvent(Message_InitiateConnection message)
    {
        var data = message.password.Bytes();
        eventID = data.Integer(0);
        messageID = data.Integer(4);
    }
}

class Message_EventExecuted
{
    public Message_InitiateConnection Message
    {
        get
        {
            var data = new List<byte>();
            data.AddRange(messageID.Bytes());
            data.AddRange(extra.Bytes());
            return new Message_InitiateConnection(MessageType.MessageID, MessageType.Executed, data.String());
        }
    }
    public int MessageID { get { return messageID; } }
    public int ExtraData { get { return extra; } }
    int messageID;
    int extra;
    public Message_EventExecuted(int messageId, int extraData = -1)
    {
        messageID = messageId;
        extra = extraData;
    }
    public Message_EventExecuted(Message_InitiateConnection message)
    {
        var data = message.password.Bytes();
        messageID = data.Integer(0);
        extra = data.Integer(4);
    }
}

class Message_Wind
{
    public Message_InitiateConnection Message
    {
        get
        {
            var data = new List<byte>();
            data.AddRange(angle.Bytes());
            data.AddRange(time.Bytes());
            return new Message_InitiateConnection(MessageType.MessageID, MessageType.Wind, data.String());
        }
    }
    public float Angle { get { return angle; } }
    public float Time { get { return time; } }
    public float angle;
    public float time;
    public Message_Wind(float Angle, float Time)
    {
        angle = Angle;
        time = Time;
    }
    public Message_Wind(Message_InitiateConnection message)
    {
        var data = message.password.Bytes();
        angle = data.Float(0);
        time = data.Float(4);
    }
}

class Message_Upgrade
{
    public Message_InitiateConnection Message
    {
        get
        {
            return new Message_InitiateConnection(MessageType.MessageID, MessageType.Upgrade, objectIndex.Bytes().String());
        }
    }
    public uint ObjectIndex { get { return objectIndex; } }
    uint objectIndex;
    public Message_Upgrade(uint ObjectIndex)
    {
        objectIndex = ObjectIndex;
    }
    public Message_Upgrade(Message_InitiateConnection message)
    {
        var data = message.password.Bytes();
        objectIndex = data.UInteger(0);
    }
}

[HarmonyPatch(typeof(Helper), "IsValidWaterTarget")]
public class Patch_SharkPlayerTargeter
{
    static bool Prefix(ref bool __result)
    {
        if (CrowdControl.sharkPacify > 0 && Environment.StackTrace.Contains("at AI_StateMachine_Shark"))
        {
            __result = false;
            return false;
        }
        return true;
    }
}

[HarmonyPatch(typeof(AI_NetworkBehaviour), "OnDestroy")]
public class Patch_AnimalDestroy
{
    static void Prefix(ref AI_NetworkBehaviour __instance)
    {
        for (int i = 0; i < CrowdControl.spawnedAnimals.Count; i++)
            if (CrowdControl.spawnedAnimals[i].animalId == __instance.ObjectIndex)
            {
                CrowdControl.spawnedAnimals.RemoveAt(i);
                break;
            }
    }
}

[HarmonyPatch(typeof(Block_Foundation), "Reinforced", MethodType.Getter)]
public class Patch_SharkBlockTargeter
{
    static void Postfix(ref bool __result)
    {
        if (CrowdControl.sharkPacify > 0 && Environment.StackTrace.Contains("at AI_StateMachine_Shark"))
            __result = true;
    }
}

[HarmonyPatch(typeof(Raft), "FixedUpdate")]
public class Patch_RaftMovement
{
    static bool Prefix(ref Raft __instance)
    {
        if (CrowdControl.windDirection.w == 0)
            return true;
        if (!Semih_Network.IsHost)
        {
            return false;
        }
        if (GameModeValueManager.GetCurrentGameModeValue().raftSpecificVariables.isRaftAlwaysAnchored)
        {
            return false;
        }
        if (!Semih_Network.WorldHasBeenRecieved && !GameManager.IsInNewGame)
        {
            return false;
        }
        Rigidbody body = Traverse.Create(__instance).Field("body").GetValue<Rigidbody>();
        if (!__instance.IsAnchored)
        {
            float speed = Traverse.Create(__instance).Field("speed").GetValue<float>();
            if (speed != 0f)
            {
                Traverse<Vector3> moveDirection = Traverse.Create(__instance).Field<Vector3>("moveDirection");
                moveDirection.Value = new Vector3(CrowdControl.windDirection.x, CrowdControl.windDirection.y, CrowdControl.windDirection.z);
                bool flag = RaftVelocityManager.MotorDirection != Vector3.zero;
                if (!flag)
                {
                    List<Sail> allSails = Sail.AllSails;
                    Vector3 vector = Vector3.zero;
                    for (int i = 0; i < allSails.Count; i++)
                    {
                        Sail sail = allSails[i];
                        if (sail.open)
                        {
                            vector += sail.GetNormalizedDirection();
                        }
                    }
                    if (vector.z < 0f)
                    {
                        if ((double)Mathf.Abs(vector.x) > 0.7)
                        {
                            vector.z = moveDirection.Value.z;
                            moveDirection.Value = new Vector3(moveDirection.Value.x, moveDirection.Value.y, 0);
                        }
                        else
                        {
                            vector.z = -0.8f;
                        }
                    }
                    moveDirection.Value += vector;
                }
                else
                {
                    moveDirection.Value = RaftVelocityManager.MotorDirection;
                }
                Traverse<float> currentMovementSpeed = Traverse.Create(__instance).Field<float>("currentMovementSpeed");
                currentMovementSpeed.Value = speed;
                if (flag)
                {
                    currentMovementSpeed.Value = RaftVelocityManager.motorSpeed;
                    if (RaftVelocityManager.MotorWheelWeightStrength == MotorWheel.WeightStrength.Weak)
                    {
                        currentMovementSpeed.Value *= 0.5f;
                    }
                    if (currentMovementSpeed.Value < speed)
                    {
                        currentMovementSpeed.Value = speed;
                    }
                }
                if (speed != 0f)
                {
                    if (currentMovementSpeed.Value > __instance.maxSpeed)
                    {
                        currentMovementSpeed.Value = __instance.maxSpeed;
                    }
                    moveDirection.Value = Vector3.ClampMagnitude(moveDirection.Value, 1f);
                    body.AddForce(moveDirection.Value * currentMovementSpeed.Value);
                }
            }
            List<SteeringWheel> steeringWheels = RaftVelocityManager.steeringWheels;
            float num = 0f;
            foreach (SteeringWheel steeringWheel in steeringWheels)
            {
                num += steeringWheel.SteeringRotation;
            }
            num = Mathf.Clamp(num, -1f, 1f);
            if (num != 0f)
            {
                Vector3 torque = new Vector3(0f, Mathf.Tan(0.017453292f * num), 0f) * __instance.maxSteeringTorque;
                body.AddTorque(torque, ForceMode.Acceleration);
            }
        }
        else
        {
            Traverse<Vector3> anchorPosition = Traverse.Create(__instance).Field<Vector3>("anchorPosition");
            float maxDistanceFromAnchorPoint = Traverse.Create(__instance).Field("maxDistanceFromAnchorPoint").GetValue<float>();
            float num2 = __instance.transform.position.DistanceXZ(anchorPosition.Value);
            if (num2 >= maxDistanceFromAnchorPoint * 3f)
            {
                anchorPosition.Value = __instance.transform.position;
            }
            if (num2 > maxDistanceFromAnchorPoint)
            {
                Vector3 vector2 = anchorPosition.Value - __instance.transform.position;
                vector2.y = 0f;
                body.AddForce(vector2.normalized * 2f);
            }
        }
        if (body.velocity.sqrMagnitude > __instance.maxVelocity)
        {
            body.velocity = Vector3.ClampMagnitude(body.velocity, __instance.maxVelocity);
        }
        Traverse.Create(__instance).Field("eventEmitter_idle").GetValue<FMODUnity.StudioEventEmitter>().SetParameter("velocity", body.velocity.sqrMagnitude / __instance.maxVelocity);
        Traverse.Create(__instance).Field("previousPosition").SetValue(body.transform.position);
        return false;
    }
}

[HarmonyPatch(typeof(Streamer), "Update")]
public class Patch_StreamerUpdate
{
    static bool Prefix(ref Streamer __instance)
    {
        if (CrowdControl.windDirection.w == 0)
            return true;
        Transform streamerTransform = Traverse.Create(__instance).Field("streamerTransform").GetValue<Transform>();
        Quaternion b = Quaternion.LookRotation(-new Vector3(CrowdControl.windDirection.x, CrowdControl.windDirection.y, CrowdControl.windDirection.z), streamerTransform.up);
        streamerTransform.rotation = Quaternion.Lerp(streamerTransform.rotation, b, Time.deltaTime * Traverse.Create(__instance).Field("rotationSpeed").GetValue<float>());
        return false;
    }
}

public class Effect
{
    public string id;
    private Action execute;
    private Func<bool> ready;
    public bool networked { get; private set; }
    public Effect(string ID, Action OnRun, Func<bool> IsReady, bool WaitForNetwork)
    {
        id = ID;
        execute = OnRun;
        ready = IsReady;
        networked = WaitForNetwork;
    }
    public bool IsReady
    {
        get
        {
            return ready();
        }
    }
    public bool Execute()
    {
        execute();
        return networked && !Semih_Network.IsHost;
    }
}

public class EffectList : List<Effect>
{
    public Effect this[string id]
    {
        get
        {
            foreach (Effect effect in this)
                if (effect.id == id)
                    return effect;
            return null;
        }
    }

    public int IndexOf(string id)
    {
        for (int i = 0; i < this.Count; i++)
            if (this[i].id == id)
                return i;
        return -1;
    }
}

public class TimedEffect
{
    public int id;
    public float time;
    public Effect reverseEffect;
    public TimedEffect(int ID, Effect reverse)
    {
        id = ID;
        reverseEffect = reverse;
        time = 0;
    }
}

public class AnimalEffect
{
    public int id;
    int _animalId;
    public string name;
    public NameTag nameTag;
    public int animalId
    {
        get
        {
            return _animalId;
        }
        set
        {
            if (value == -1)
                _animalId = -1;
            else
            {
                foreach (AI_NetworkBehaviour entity in GameObject.FindObjectsOfType<AI_NetworkBehaviour>())
                    if (entity.ObjectIndex == value)
                        nameTag = new NameTag(entity.networkEntity, name);
                if (nameTag == null)
                    _animalId = -1;
                else
                    _animalId = value;
            }
        }
    }
    public AnimalEffect(int ID, string Name, int AnimalID = -1)
    {
        id = ID;
        name = Name;
        animalId = AnimalID;
    }
}

public class NameTag
{
    private Canvas canvas;
    Network_Entity parent;
    GameObject container;

    public NameTag(Network_Entity creature, string name)
    {
        parent = creature;
        container = new GameObject();
        container.transform.SetParent(parent.transform, false);
        container.transform.localPosition = new Vector3(0, 2.25f, 0);
        RectTransform rect = container.AddComponent<RectTransform>();
        rect.sizeDelta = new Vector2(1, 1f);
        canvas = container.AddComponent<Canvas>();
        canvas.transform.SetParent(container.transform, false);
        canvas.transform.localScale = new Vector3(0.001f, 0.001f, 0.001f);
        RectTransform trans = canvas.GetComponent<RectTransform>();
        trans.sizeDelta = new Vector2(1, 1f);
        Text numDisplay = CrowdControl.CreateText(container.transform, 0, 0, name, 300, ComponentManager<CanvasHelper>.Value.dropText.color, 3000, 450, ComponentManager<CanvasHelper>.Value.dropText.font).GetComponent<Text>();
        numDisplay.rectTransform.sizeDelta = new Vector2(numDisplay.preferredWidth, numDisplay.preferredHeight);
        CrowdControl.CopyTextShadow(numDisplay.gameObject, ComponentManager<CanvasHelper>.Value.dropText.gameObject);
    }

    public void UpdateFacing()
    {
        Vector3 dirVec = Helper.MainCamera.transform.position - canvas.transform.position;
        float angle = Mathf.Acos(dirVec.z / dirVec.XZOnly().magnitude) / (float)Math.PI * 180f + 180;
        container.transform.rotation = Quaternion.Euler(0, dirVec.x < 0 ? -angle : angle, 0);
    }
}

public enum RequestType
{
    Test = 0,
    Start = 1,
    Stop = 2
}

public class Request
{
    public uint id;
    public string code;
    public object[] parameters;
    public string viewer;
    public int? cost;
    public RequestType type;

}

public class Response
{
    public uint id;
    public EffectResult status;
    public string message;
    public ResponseType type = ResponseType.EffectRequest;

    public enum ResponseType : byte
    {
        EffectRequest = 0,
        KeepAlive = 255
    }
}

public enum EffectResult
{
    /// <summary>The effect executed successfully.</summary>
    Success = 0,

    /// <summary>The effect failed to trigger, but is still available for use. Viewer(s) will be refunded.</summary>
    Failure = 1,

    /// <summary>Same as "Failure" but the effect is no longer available for use.</summary>
    Unavailable = 2,

    /// <summary>The effect cannot be triggered right now, try again in a few seconds.</summary>
    Retry = 3
}

public class AngleRange
{
    List<Range> ranges = new List<Range>() { new Range(0, Angle.max) };
    public AngleRange(double Min, double Max)
    {
        ranges = new List<Range>() { new Range(Min, Max) };
    }
    public AngleRange() { }
    AngleRange(List<Range> Ranges)
    {
        ranges = Ranges;
    }

    public bool Empty
    {
        get
        {
            return ranges.Count == 0;
        }
    }

    public static AngleRange operator -(AngleRange a, AngleRange b)
    {
        List<Range> tmp = new List<Range>();
        foreach (Range range in b.ranges)
            tmp.AddRange(a - range);
        return new AngleRange(tmp);
    }

    public static List<Range> operator -(AngleRange a, Range b)
    {
        List<Range> tmp = new List<Range>();
        foreach (Range range in a.ranges)
            tmp.AddRange(range - b);
        return tmp;
    }

    public struct Angle
    {
        public const double max = 2 * Math.PI;
        double value;
        public Angle(double angle)
        {
            value = angle % max + (angle < 0 ? max : 0);
            if (value == 0 && angle > 0)
                value = max;
        }

        public static implicit operator double(Angle a) => a.value;
        public static implicit operator Angle(double a) => new Angle(a);
    }
    public struct Range
    {
        Angle min;
        public Angle Size;

        public Range(double Min, double Max)
        {
            min = Min;
            Size = Max - min;
        }
        public Angle Min
        {
            get { return min; }
            set
            {
                Size += min - value;
                min = value;
            }
        }
        public Angle Max
        {
            get { return min + Size; }
            set
            {
                Size = value - min;
            }
        }
        public double dMax
        {
            get { return min + Size; }
        }
        public static List<Range> operator -(Range a, Range b)
        {
            List<Range> tmp = new List<Range>();
            if (a.Size == Angle.max)
            {
                tmp.Add(new Range(b.Max, b.min));
                return tmp;
            }
            if (b.Size == Angle.max)
                return tmp;
            bool flag1 = a.Contains(b.min);
            bool flag2 = a.Contains(b.Max);
            bool flag3 = b.Contains(a.min);
            bool flag4 = b.Contains(a.Max);
            if (flag3 && flag4 && flag1 == flag2)
            {
                if (flag1)
                    tmp.Add(new Range(b.Max, b.min));
                return tmp;
            }
            if (flag3)
                tmp.Add(new Range(b.Max, a.Max));
            else if (flag1)
            {
                tmp.Add(new Range(a.min, b.min));
                if (flag2)
                    tmp.Add(new Range(b.Max, a.Max));
            }
            else
                tmp.Add(a);
            return tmp;
        }

        public bool Contains(Angle value)
        {
            return dMax > Angle.max ? value > min || value < Max : value > min && value < dMax;
        }
    }
}